from abstract_utilities import make_list
import json,pprint
from .price_utils import *
from .signature_data_parse import get_log_messages_from_txn,get_instructions_from_txn,get_inner_instructions_from_txn
from .account_key_utils import get_all_account_keys
from .constants import TOKEN_PROGRAM_ID
def ifListGetSection(list_obj,i=0):
    if list_obj and isinstance(list_obj,list) and len(list_obj)>i:
        list_obj = list_obj[i]
    return list_obj
def get_logs_from_index(txnData,index=None):
    if index is not None:
        allLogs = get_log_messages_from_txn(txnData)
        endLog = get_end_log_index(txnData,index)
        return allLogs[index:endLog]
def get_program_ids(txnData):
    allLogs = get_log_messages_from_txn(txnData)
    return [log.split(' ')[1] for log in allLogs if 'invoke' in log]
def get_program_id_from_log(logs):
    for log in make_list(logs):
        if 'invoke' in log.lower():
            return log.split(' ')[1]
def get_stack_height_from_logs(logs):
    for log in make_list(logs):
        if 'invoke' in log.lower():
            return int(log.split(' ')[-1][1:-1])
def get_end_log_index(txnData,index):
    allLogs = get_log_messages_from_txn(txnData)
    i=0
    for i,log in enumerate(allLogs[index+1:]):
        if 'invoke' in log.lower():
            return index+1+i
    return len(allLogs)
def get_stack_height_from_logs(logs):
    for log in make_list(logs):
        if 'invoke' in log.lower():
            return int(log.split(' ')[-1][1:-1])
def get_program_id_from_log(logs):
    for log in make_list(logs):
        if 'invoke' in log.lower():
            return log.split(' ')[1]

def get_all_logs(txnData):
    logits = []
    allLogs = get_log_messages_from_txn(txnData)
    for i,log in enumerate(allLogs):
        if 'invoke' in log.lower():
            logits.append([])
        logits[-1].append(log)
    start = 0
    for i,log in enumerate(logits):    
        length = len(log)
        end = start+length
        logits[i]={"programId":get_program_id_from_log(log[0]),
                   "start":start,
                   "end":end-1,
                   'stackHeight':get_stack_height_from_logs(log[0]) or 1,
                   'events':[event.split(':')[-1] or 'unknown' for event in get_log_events(log)],
                   'vars':[event.split(':')[1:] or 'unknown' for event in get_log_events(log)],
                   'logs':log}
        start = end
    return logits
def split_log_for_instruction(log):
    return log.split('log:')[-1].split('Instruction:')[-1]
def clean_split_string(string,delim=' '):
    return delim.join([spl for spl in string.split(' ') if spl])
def get_log_events(logs=None,index=None):
    return [clean_split_string(split_log_for_instruction(log)) for log in logs if 'log:' in log.lower() or 'instruction:' in log.lower()]
def get_instruction_accounts(instruction,txnData):
    accounts = get_all_account_keys(txnData)
    return [accounts[index] for index in instruction.get('accounts',[]) if index < len(accounts)]  
def get_instruction_info(instruction,txnData,instruction_index=0):
  
    stack_height = instruction.get('stackHeight') or 1
    accounts = instruction.get('accounts',[])
    associatedAccounts = get_instruction_accounts(instruction,txnData)
    instruction_info = {'instruction_index':instruction_index,'accounts':accounts,'associatedAccounts':associatedAccounts,'instructionStackHeight':stack_height}

    instruction_info.update(get_for_program_ids_info(txnData)[instruction_index])
    return instruction_info
def process_instructions(instructions,txnData,start_index=1):
    catalog = []
    for i, inst in enumerate(instructions):
        instruction_index = start_index-1 + i
        instruction_info = get_instruction_info(inst,txnData,instruction_index)
        catalog.append(instruction_info)
    return catalog
def get_instructions_catalog(txnData,printIt=False,saveIt=False):
    instructions = get_instructions_from_txn(txnData)
    outer_instructions_catalog = process_instructions(instructions,txnData)
    inner_instructions = get_inner_instructions_from_txn(txnData)
    if inner_instructions:
        inner_instructions_catalog = process_instructions(inner_instructions,txnData,start_index=len(instructions))
    complete_catalog =  outer_instructions_catalog+inner_instructions_catalog
    if printIt:
        pprint.pprint(complete_catalog)
    if saveIt:
        with open(saveIt, 'w') as f:
            json.dump(complete_catalog, f, indent=4)
    return complete_catalog

def find_in_catalog(string,txnData,programId=None):
    complete_catalog = get_instructions_catalog(txnData)
    return [txn for txn in complete_catalog if [event for event in txn['events'] if string.lower() in event.lower()]]
def findKeyValueIdInCatalog(key,value,txnData,programId=None):
    complete_catalog = get_instructions_catalog(txnData)
    if programId:
        complete_catalog = findKeyValueIdInCatalog('programId',programId,txnData)
    return [txn for txn in make_list(complete_catalog) if txn.get(key) == value]
def find_account_in_catalog(account,catalog):
    return ifListGetSection([txn for txn in make_list(catalog) if account in txn.get('associatedAccounts')])
def associate_logs_with_instructions(txnData):
    accountKeys = get_all_account_keys(txnData)
    instructions = txnData['transaction']['message']['instructions']
    innerInstructions = txnData['meta']['innerInstructions'][0]['instructions']
    allLogs = txnData['meta']['logMessages']
    for logIndex,log in enumerate(allLogs):
        log_programId = log['programId']
        log_stackHeight = log.get('stackHeight')  # Default to 0 if stackHeight is missing
        # Search for matching instructions by programId and stackHeight
        for instIndex,allInstruction in enumerate([instructions,innerInstructions]):
            for i,instruction in enumerate(allInstruction):
                program_id_index = instruction.get('programIdIndex')
                if program_id_index is not None:
                    instruction_program_id = accountKeys[program_id_index]
                    instruction_stack_height = instruction.get('stackHeight', 1)
                    if instruction_program_id == log_programId and instruction_stack_height == log_stackHeight:
                        # Add log data to the matching instruction
                        instruction.update(log)
                        instruction['associatedAccounts'] = [accountKeys[index] for index in instruction['accounts']]
                        if instIndex == 0:
                            instructions[i] = instruction
                        else:
                            innerInstructions[i] = instruction
                        allLogs[logIndex].update(instruction)
    txnData['transaction']['message']['instructions'] = instructions
    txnData['meta']['innerInstructions'][0]['instructions'] = innerInstructions
    txnData['meta']['logMessages'] = allLogs
    return txnData
def update_initial_txnData(txnData):
    accountKeys = get_all_account_keys(txnData)
    txnData = update_balance_data(txnData)
    txnData['transaction']['message']['instructions'] = [{**inst,"instructionIndex":instIndex,"programId":accountKeys[inst.get('programIdIndex')],"stackHeight":inst.get('stackHeight', 1),"associatedAccounts":[accountKeys[index] for index in inst['accounts']]} for instIndex,inst in enumerate(txnData['transaction']['message']['instructions'])]
    txnData['meta']['innerInstructions'][0]['instructions'] = [{**inst,"instructionIndex":instIndex+len(txnData['transaction']['message']['instructions']),"programId":accountKeys[inst.get('programIdIndex')],"stackHeight":inst.get('stackHeight', 1),"associatedAccounts":[accountKeys[index] for index in inst['accounts']]} for instIndex,inst in enumerate(txnData['meta']['innerInstructions'][0]['instructions'])]
    txnData['meta']['logMessages'] = get_all_logs(txnData)
    return txnData
def get_for_program_ids_info(txnData):
    txnData = update_initial_txnData(txnData)
    txnData = associate_logs_with_instructions(txnData)
    return txnData



# Predefined table configurations
def get_insert_list():
    return [{'fresh_call':True,'tableName': 'isconnected', 'columnSearch': None, 'insertName': 'connected', 'type': 'bool', 
                        'searchQuery': 'SELECT connected FROM isconnected WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO isconnected (None, connected) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS isconnected (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,connected JSONB NOT NULL);'},
                       {'fresh_call':True,'tableName': 'getbalance', 'columnSearch': 'pubkey', 'insertName': 'balance', 'type': 'Balance', 
                        'searchQuery': 'SELECT balance FROM getbalance WHERE pubkey = %s;',
                        'insertQuery': 'INSERT INTO getbalance (pubkey, balance) VALUES (%s, %s)',
                        'table': 'CREATE TABLE IF NOT EXISTS getbalance (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,balance JSONB NOT NULL);'},
                       {'fresh_call':True,'tableName': 'getaccountinfo', 'columnSearch': 'pubkey', 'insertName': 'account_info', 'type': 'AccountInfo',
                        'searchQuery': 'SELECT account_info FROM getaccountinfo WHERE pubkey = %s;',
                        'insertQuery': 'INSERT INTO getaccountinfo (pubkey, account_info) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getaccountinfo (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,account_info JSONB NOT NULL);'},
                       {'fresh_call':True,'tableName': 'getaccountinfojsonparsed', 'columnSearch': 'pubkey', 'insertName': 'account_info_json_parsed', 'type': 'AccountInfoMaybeJsonParsed', 
                        'searchQuery': 'SELECT account_info_json_parsed FROM getaccountinfojsonparsed WHERE pubkey = %s;',
                        'insertQuery': 'INSERT INTO getaccountinfojsonparsed (pubkey, account_info_json_parsed) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getaccountinfojsonparsed (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,account_info_json_parsed JSONB NOT NULL);'},
                       {'fresh_call':True,'tableName': 'getblockcommitment', 'columnSearch': 'slot', 'insertName': 'block_commitment', 'type': 'BlockCommitment', 
                        'searchQuery': 'SELECT block_commitment FROM getblockcommitment WHERE slot = %s;',
                        'insertQuery': 'INSERT INTO getblockcommitment (slot, block_commitment) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getblockcommitment (id SERIAL PRIMARY KEY,slot REAL UNIQUE NOT NULL,block_commitment JSONB NOT NULL);'},
                        {'fresh_call':True,'tableName': 'getblocktime', 'columnSearch': 'slot', 'insertName': 'block_time', 'type': 'BlockTime', 
                        'searchQuery': 'SELECT block_time FROM getblocktime WHERE slot = %s;', 
                        'insertQuery': 'INSERT INTO getblocktime (slot, block_time) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getblocktime (id SERIAL PRIMARY KEY,slot REAL UNIQUE NOT NULL,block_time JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getclusternodes', 'columnSearch': None, 'insertName': 'cluster_nodes', 'type': 'ClusterNodes', 
                        'searchQuery': 'SELECT cluster_nodes FROM getclusternodes WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getclusternodes (None, cluster_nodes) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getclusternodes (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,cluster_nodes JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getblock', 'columnSearch': 'slot', 'insertName': 'block', 'type': 'Block', 
                        'searchQuery': 'SELECT block FROM getblock WHERE slot = %s;', 
                        'insertQuery': 'INSERT INTO getblock (slot, block) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getblock (id SERIAL PRIMARY KEY,slot REAL UNIQUE NOT NULL,block JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getrecentperformancesamples', 'columnSearch': 'limit', 'insertName': 'recent_performance_samples', 'type': 'RecentPerformanceSamples', 
                        'searchQuery': 'SELECT recent_performance_samples FROM getrecentperformancesamples WHERE limit = %s;', 
                        'insertQuery': 'INSERT INTO getrecentperformancesamples (limit, recent_performance_samples) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getrecentperformancesamples (id SERIAL PRIMARY KEY,limit REAL UNIQUE NOT NULL,recent_performance_samples JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getblockheight', 'columnSearch': 'commitment', 'insertName': 'block_height', 'type': 'BlockHeight', 
                        'searchQuery': 'SELECT block_height FROM getblockheight WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getblockheight (commitment, block_height) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getblockheight (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,block_height JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getblocks', 'columnSearch': 'start_slot', 'insertName': 'blocks', 'type': 'Blocks', 
                        'searchQuery': 'SELECT blocks FROM getblocks WHERE start_slot = %s;', 
                        'insertQuery': 'INSERT INTO getblocks (start_slot, blocks) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getblocks (id SERIAL PRIMARY KEY,start_slot REAL UNIQUE NOT NULL,blocks JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getsignaturesforaddress', 'columnSearch': 'account', 'insertName': 'signatures_for_address', 'type': 'SignaturesForAddress', 
                        'searchQuery': 'SELECT signatures_for_address FROM getsignaturesforaddress WHERE account = %s;', 
                        'insertQuery': 'INSERT INTO getsignaturesforaddress (account, signatures_for_address) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getsignaturesforaddress (id SERIAL PRIMARY KEY,account REAL UNIQUE NOT NULL,signatures_for_address JSONB NOT NULL);'}, 
                        {'fresh_call':False,'tableName': 'gettransaction', 'columnSearch': 'signature', 'insertName': 'transaction', 'type': 'Transaction', 
                        'searchQuery': 'SELECT transaction FROM gettransaction WHERE signature = %s;', 
                        'insertQuery': 'INSERT INTO gettransaction (signature, transaction) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettransaction (id SERIAL PRIMARY KEY,signature REAL UNIQUE NOT NULL,transaction JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getepochinfo', 'columnSearch': 'commitment', 'insertName': 'epoch_info', 'type': 'EpochInfo', 
                        'searchQuery': 'SELECT epoch_info FROM getepochinfo WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getepochinfo (commitment, epoch_info) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getepochinfo (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,epoch_info JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getepochschedule', 'columnSearch': None, 'insertName': 'epoch_schedule', 'type': 'EpochSchedule', 
                        'searchQuery': 'SELECT epoch_schedule FROM getepochschedule WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getepochschedule (None, epoch_schedule) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getepochschedule (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,epoch_schedule JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getfeeformessage', 'columnSearch': 'message', 'insertName': 'fee_for_message', 'type': 'FeeForMessage', 
                        'searchQuery': 'SELECT fee_for_message FROM getfeeformessage WHERE message = %s;', 
                        'insertQuery': 'INSERT INTO getfeeformessage (message, fee_for_message) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getfeeformessage (id SERIAL PRIMARY KEY,message REAL UNIQUE NOT NULL,fee_for_message JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getfirstavailableblock', 'columnSearch': None, 'insertName': 'first_available_block', 'type': 'FirstAvailableBlock', 
                        'searchQuery': 'SELECT first_available_block FROM getfirstavailableblock WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getfirstavailableblock (None, first_available_block) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getfirstavailableblock (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,first_available_block JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getgenesishash', 'columnSearch': None, 'insertName': 'genesis_hash', 'type': 'GenesisHash', 
                        'searchQuery': 'SELECT genesis_hash FROM getgenesishash WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getgenesishash (None, genesis_hash) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getgenesishash (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,genesis_hash JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getidentity', 'columnSearch': None, 'insertName': 'identity', 'type': 'Identity', 
                        'searchQuery': 'SELECT identity FROM getidentity WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getidentity (None, identity) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getidentity (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,identity JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getinflationgovernor', 'columnSearch': 'commitment', 'insertName': 'inflation_governor', 'type': 'InflationGovernor', 
                        'searchQuery': 'SELECT inflation_governor FROM getinflationgovernor WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getinflationgovernor (commitment, inflation_governor) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getinflationgovernor (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,inflation_governor JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getinflationrate', 'columnSearch': None, 'insertName': 'inflation_rate', 'type': 'InflationRate', 
                        'searchQuery': 'SELECT inflation_rate FROM getinflationrate WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getinflationrate (None, inflation_rate) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getinflationrate (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,inflation_rate JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getinflationreward', 'columnSearch': 'pubkeys', 'insertName': 'inflation_reward', 'type': 'InflationReward', 
                        'searchQuery': 'SELECT inflation_reward FROM getinflationreward WHERE pubkeys = %s;', 
                        'insertQuery': 'INSERT INTO getinflationreward (pubkeys, inflation_reward) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getinflationreward (id SERIAL PRIMARY KEY,pubkeys REAL UNIQUE NOT NULL,inflation_reward JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getlargestaccounts', 'columnSearch': 'filter_opt', 'insertName': 'largest_accounts', 'type': 'LargestAccounts', 
                        'searchQuery': 'SELECT largest_accounts FROM getlargestaccounts WHERE filter_opt = %s;', 
                        'insertQuery': 'INSERT INTO getlargestaccounts (filter_opt, largest_accounts) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getlargestaccounts (id SERIAL PRIMARY KEY,filter_opt REAL UNIQUE NOT NULL,largest_accounts JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getleaderschedule', 'columnSearch': 'epoch', 'insertName': 'leader_schedule', 'type': 'LeaderSchedule', 
                        'searchQuery': 'SELECT leader_schedule FROM getleaderschedule WHERE epoch = %s;', 
                        'insertQuery': 'INSERT INTO getleaderschedule (epoch, leader_schedule) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getleaderschedule (id SERIAL PRIMARY KEY,epoch REAL UNIQUE NOT NULL,leader_schedule JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getminimumbalanceforrentexemption', 'columnSearch': 'usize', 'insertName': 'minimum_balance_for_rent_exemption', 'type': 'MinimumBalanceForRentExemption', 
                        'searchQuery': 'SELECT minimum_balance_for_rent_exemption FROM getminimumbalanceforrentexemption WHERE usize = %s;', 
                        'insertQuery': 'INSERT INTO getminimumbalanceforrentexemption (usize, minimum_balance_for_rent_exemption) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getminimumbalanceforrentexemption (id SERIAL PRIMARY KEY,usize REAL UNIQUE NOT NULL,minimum_balance_for_rent_exemption JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getmultipleaccounts', 'columnSearch': 'pubkeys', 'insertName': 'multiple_accounts', 'type': 'MultipleAccounts', 
                        'searchQuery': 'SELECT multiple_accounts FROM getmultipleaccounts WHERE pubkeys = %s;', 
                        'insertQuery': 'INSERT INTO getmultipleaccounts (pubkeys, multiple_accounts) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getmultipleaccounts (id SERIAL PRIMARY KEY,pubkeys REAL UNIQUE NOT NULL,multiple_accounts JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getmultipleaccountsjsonparsed', 'columnSearch': 'pubkeys', 'insertName': 'multiple_accounts_json_parsed', 'type': 'MultipleAccountsMaybeJsonParsed', 
                        'searchQuery': 'SELECT multiple_accounts_json_parsed FROM getmultipleaccountsjsonparsed WHERE pubkeys = %s;', 
                        'insertQuery': 'INSERT INTO getmultipleaccountsjsonparsed (pubkeys, multiple_accounts_json_parsed) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getmultipleaccountsjsonparsed (id SERIAL PRIMARY KEY,pubkeys REAL UNIQUE NOT NULL,multiple_accounts_json_parsed JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getprogramaccounts', 'columnSearch': 'pubkey', 'insertName': 'program_accounts', 'type': 'ProgramAccounts', 
                        'searchQuery': 'SELECT program_accounts FROM getprogramaccounts WHERE pubkey = %s;', 
                        'insertQuery': 'INSERT INTO getprogramaccounts (pubkey, program_accounts) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getprogramaccounts (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,program_accounts JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getprogramaccountsjsonparsed', 'columnSearch': 'pubkey', 'insertName': 'program_accounts_json_parsed', 'type': 'ProgramAccountsMaybeJsonParsed', 
                        'searchQuery': 'SELECT program_accounts_json_parsed FROM getprogramaccountsjsonparsed WHERE pubkey = %s;', 
                        'insertQuery': 'INSERT INTO getprogramaccountsjsonparsed (pubkey, program_accounts_json_parsed) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getprogramaccountsjsonparsed (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,program_accounts_json_parsed JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getlatestblockhash', 'columnSearch': 'commitment', 'insertName': 'latest_blockhash', 'type': 'LatestBlockhash', 
                        'searchQuery': 'SELECT latest_blockhash FROM getlatestblockhash WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getlatestblockhash (commitment, latest_blockhash) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getlatestblockhash (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,latest_blockhash JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getsignaturestatuses', 'columnSearch': 'signatures', 'insertName': 'signature_statuses', 'type': 'SignatureStatuses', 
                        'searchQuery': 'SELECT signature_statuses FROM getsignaturestatuses WHERE signatures = %s;', 
                        'insertQuery': 'INSERT INTO getsignaturestatuses (signatures, signature_statuses) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getsignaturestatuses (id SERIAL PRIMARY KEY,signatures REAL UNIQUE NOT NULL,signature_statuses JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getslot', 'columnSearch': 'commitment', 'insertName': 'slot', 'type': 'Slot', 
                        'searchQuery': 'SELECT slot FROM getslot WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getslot (commitment, slot) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getslot (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,slot JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getslotleader', 'columnSearch': 'commitment', 'insertName': 'slot_leader', 'type': 'SlotLeader', 
                        'searchQuery': 'SELECT slot_leader FROM getslotleader WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getslotleader (commitment, slot_leader) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getslotleader (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,slot_leader JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getstakeactivation', 'columnSearch': 'pubkey', 'insertName': 'stake_activation', 'type': 'StakeActivation', 
                        'searchQuery': 'SELECT stake_activation FROM getstakeactivation WHERE pubkey = %s;', 
                        'insertQuery': 'INSERT INTO getstakeactivation (pubkey, stake_activation) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getstakeactivation (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,stake_activation JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getsupply', 'columnSearch': 'commitment', 'insertName': 'supply', 'type': 'Supply', 
                        'searchQuery': 'SELECT supply FROM getsupply WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getsupply (commitment, supply) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getsupply (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,supply JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettokenaccountbalance', 'columnSearch': 'pubkey', 'insertName': 'token_account_balance', 'type': 'TokenAccountBalance', 
                        'searchQuery': 'SELECT token_account_balance FROM gettokenaccountbalance WHERE pubkey = %s;', 
                        'insertQuery': 'INSERT INTO gettokenaccountbalance (pubkey, token_account_balance) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettokenaccountbalance (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,token_account_balance JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettokenaccountsbydelegate', 'columnSearch': 'delegate', 'insertName': 'token_accounts_by_delegate', 'type': 'TokenAccountsByDelegate', 
                        'searchQuery': 'SELECT token_accounts_by_delegate FROM gettokenaccountsbydelegate WHERE delegate = %s;', 
                        'insertQuery': 'INSERT INTO gettokenaccountsbydelegate (delegate, token_accounts_by_delegate) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettokenaccountsbydelegate (id SERIAL PRIMARY KEY,delegate REAL UNIQUE NOT NULL,token_accounts_by_delegate JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettokenaccountsbydelegatejsonparsed', 'columnSearch': 'delegate', 'insertName': 'token_accounts_by_delegate_json_parsed', 'type': 'TokenAccountsByDelegateJsonParsed', 
                        'searchQuery': 'SELECT token_accounts_by_delegate_json_parsed FROM gettokenaccountsbydelegatejsonparsed WHERE delegate = %s;', 
                        'insertQuery': 'INSERT INTO gettokenaccountsbydelegatejsonparsed (delegate, token_accounts_by_delegate_json_parsed) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettokenaccountsbydelegatejsonparsed (id SERIAL PRIMARY KEY,delegate REAL UNIQUE NOT NULL,token_accounts_by_delegate_json_parsed JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettokenaccountsbyowner', 'columnSearch': 'owner', 'insertName': 'token_accounts_by_owner', 'type': 'TokenAccountsByOwner', 
                        'searchQuery': 'SELECT token_accounts_by_owner FROM gettokenaccountsbyowner WHERE owner = %s;', 
                        'insertQuery': 'INSERT INTO gettokenaccountsbyowner (owner, token_accounts_by_owner) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettokenaccountsbyowner (id SERIAL PRIMARY KEY,owner REAL UNIQUE NOT NULL,token_accounts_by_owner JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettokenaccountsbyownerjsonparsed', 'columnSearch': 'owner', 'insertName': 'token_accounts_by_owner_json_parsed', 'type': 'TokenAccountsByOwnerJsonParsed', 
                        'searchQuery': 'SELECT token_accounts_by_owner_json_parsed FROM gettokenaccountsbyownerjsonparsed WHERE owner = %s;', 
                        'insertQuery': 'INSERT INTO gettokenaccountsbyownerjsonparsed (owner, token_accounts_by_owner_json_parsed) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettokenaccountsbyownerjsonparsed (id SERIAL PRIMARY KEY,owner REAL UNIQUE NOT NULL,token_accounts_by_owner_json_parsed JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettokenlargestaccounts', 'columnSearch': 'pubkey', 'insertName': 'token_largest_accounts', 'type': 'TokenLargestAccounts', 
                        'searchQuery': 'SELECT token_largest_accounts FROM gettokenlargestaccounts WHERE pubkey = %s;', 
                        'insertQuery': 'INSERT INTO gettokenlargestaccounts (pubkey, token_largest_accounts) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettokenlargestaccounts (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,token_largest_accounts JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettokensupply', 'columnSearch': 'pubkey', 'insertName': 'token_supply', 'type': 'TokenSupply', 
                        'searchQuery': 'SELECT token_supply FROM gettokensupply WHERE pubkey = %s;', 
                        'insertQuery': 'INSERT INTO gettokensupply (pubkey, token_supply) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettokensupply (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,token_supply JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'gettransactioncount', 'columnSearch': 'commitment', 'insertName': 'transaction_count', 'type': 'TransactionCount', 
                        'searchQuery': 'SELECT transaction_count FROM gettransactioncount WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO gettransactioncount (commitment, transaction_count) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS gettransactioncount (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,transaction_count JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getminimumledgerslot', 'columnSearch': None, 'insertName': 'minimum_ledger_slot', 'type': 'MinimumLedgerSlot', 
                        'searchQuery': 'SELECT minimum_ledger_slot FROM getminimumledgerslot WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getminimumledgerslot (None, minimum_ledger_slot) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getminimumledgerslot (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,minimum_ledger_slot JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getversion', 'columnSearch': None, 'insertName': 'version', 'type': 'Version', 
                        'searchQuery': 'SELECT version FROM getversion WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO getversion (None, version) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getversion (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,version JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'getvoteaccounts', 'columnSearch': 'commitment', 'insertName': 'vote_accounts', 'type': 'VoteAccounts', 
                        'searchQuery': 'SELECT vote_accounts FROM getvoteaccounts WHERE commitment = %s;', 
                        'insertQuery': 'INSERT INTO getvoteaccounts (commitment, vote_accounts) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS getvoteaccounts (id SERIAL PRIMARY KEY,commitment REAL UNIQUE NOT NULL,vote_accounts JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'requestairdrop', 'columnSearch': 'pubkey', 'insertName': 'airdrop', 'type': 'RequestAirdrop', 
                        'searchQuery': 'SELECT airdrop FROM requestairdrop WHERE pubkey = %s;', 
                        'insertQuery': 'INSERT INTO requestairdrop (pubkey, airdrop) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS requestairdrop (id SERIAL PRIMARY KEY,pubkey REAL UNIQUE NOT NULL,airdrop JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'sendrawtransaction', 'columnSearch': 'txn', 'insertName': 'raw_transaction', 'type': 'SendTransaction', 
                        'searchQuery': 'SELECT raw_transaction FROM sendrawtransaction WHERE txn = %s;', 
                        'insertQuery': 'INSERT INTO sendrawtransaction (txn, raw_transaction) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS sendrawtransaction (id SERIAL PRIMARY KEY,txn REAL UNIQUE NOT NULL,raw_transaction JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'sendtransaction', 'columnSearch': 'txn', 'insertName': 'transaction', 'type': 'SendTransaction', 
                        'searchQuery': 'SELECT transaction FROM sendtransaction WHERE txn = %s;', 
                        'insertQuery': 'INSERT INTO sendtransaction (txn, transaction) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS sendtransaction (id SERIAL PRIMARY KEY,txn REAL UNIQUE NOT NULL,transaction JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'simulatetransaction', 'columnSearch': 'txn', 'insertName': 'transaction', 'type': 'SimulateTransaction', 
                        'searchQuery': 'SELECT transaction FROM simulatetransaction WHERE txn = %s;', 
                        'insertQuery': 'INSERT INTO simulatetransaction (txn, transaction) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS simulatetransaction (id SERIAL PRIMARY KEY,txn REAL UNIQUE NOT NULL,transaction JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'validatorexit', 'columnSearch': None, 'insertName': 'exit', 'type': 'ValidatorExit', 
                        'searchQuery': 'SELECT exit FROM validatorexit WHERE None = %s;', 
                        'insertQuery': 'INSERT INTO validatorexit (None, exit) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS validatorexit (id SERIAL PRIMARY KEY,None REAL UNIQUE NOT NULL,exit JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'postsendwithconfirm', 'columnSearch': 'resp', 'insertName': '_post_send_with_confirm', 'type': 'SendTransaction', 
                        'searchQuery': 'SELECT _post_send_with_confirm FROM postsendwithconfirm WHERE resp = %s;', 
                        'insertQuery': 'INSERT INTO postsendwithconfirm (resp, _post_send_with_confirm) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS postsendwithconfirm (id SERIAL PRIMARY KEY,resp REAL UNIQUE NOT NULL,_post_send_with_confirm JSONB NOT NULL);'}, 
                        {'fresh_call':True,'tableName': 'confirmtransaction', 'columnSearch': 'tx_sig', 'insertName': 'transaction', 'type': 'SignatureStatuses', 
                        'searchQuery': 'SELECT transaction FROM confirmtransaction WHERE tx_sig = %s;', 
                        'insertQuery': 'INSERT INTO confirmtransaction (tx_sig, transaction) VALUES (%s, %s)', 
                        'table': 'CREATE TABLE IF NOT EXISTS confirmtransaction (id SERIAL PRIMARY KEY,tx_sig REAL UNIQUE NOT NULL,transaction JSONB NOT NULL);'}
]



from solana.rpc.core import _ClientCore
from typing import Dict, List, Optional, Sequence, Union
from solana.rpc.commitment import Commitment, Finalized
from abstract_solana import * 
import inspect,asyncio,json,requests
from abstract_apis import *
from abstract_utilities import is_number
rate_limiter = RateLimiter()
def convert_to_lower(string_obj):
    return ''.join(f"_{char.lower()}" if char.isupper() else char for char in str(string_obj))
def convert_to_upper(string_obj):
    words = string_obj.split('_')
    return words[0] + ''.join(word.capitalize() for word in words[1:])
def convert_to_body(string_obj):
    return f"_{convert_to_lower(string_obj)}_body"
class Client(_ClientCore):
    def __init__(
        self,
        endpoint: Optional[str] = None,
        commitment: Optional[Commitment] = "confirmed",
        timeout: float = 10,
        extra_headers: Optional[Dict[str, str]] = None,
    ):
        """Init API client."""
        super().__init__(commitment)
client = Client()
def get_function(function_string):
    try:
        function = getattr(client, function_string)
    except:
        function =None
    return function
def call_function(function,*args,**kwargs):
    result = None
    if function:
        try:
            result = function(*args,**kwargs)
        except TypeError as e:
            print(f"Error calling function: {e}")
            result = None
    return result
def get_defaults():
    return {'limit':1000,'encoding': "jsonParsed",'usize': 1024,'preflightCommitment': 'confirmed','sig_verify': True,'commitment': 'confirmed','maxSupportedTransactionVersion': 0,'max_supported_transaction_version': 0,'skipPreflight': True,'opts': {'skipPreflight': True, 'preflightCommitment': 'confirmed'}}
def get_default_value(key):
    return get_defaults().get(key)
def convert_value(key,value):
    default_value = get_default_value(key)
    if value is None and default_value:
        return default_value
    pubkeys = ['address','account','pubkeys','pubkey','mint','owner','delegate']
    if key in pubkeys:
        if isinstance(value,list):
            value = [get_pubkey(pubkey) for pubkey in value]
        else:
            value = get_pubkey(value)
        return value
    signatures = ['until','before','tx_sig','signature','before','signatures']
    if key in signatures:
        if isinstance(value,list):
            value = [get_sigkey(signature) for signature in value]
        else:
            value = get_sigkey(value)
        return value
    return value
def get_conversions(variables,*args,**kwargs):
    for i,arg in enumerate(args):
        variable = variables[i]
        kwargs[variable] = arg
    for key,value in kwargs.items():
        if key in variables:
            kwargs[key] = convert_value(key,value)
            get_default_value(key)
    for variable in variables:
        if variable not in kwargs:
            kwargs[variable] = get_default_value(variable)
    return kwargs
def get_params(function):
    sig = inspect.signature(function)
    return list(sig.parameters.keys())
def get_rpc_dict(endpoint,*args,**kwargs):
    body_call = convert_to_body(endpoint)
    function = get_function(body_call)
    variables = get_params(function)
    kwargs = get_conversions(variables,*args,**kwargs)
    kwargs = json.loads(str(call_function(function,**kwargs)))
    return kwargs
def abstract_solana_rate_limited_call(endpoint,*args,**kwargs):
    rpc_dict = get_rpc_dict(endpoint,*args,**kwargs)
    url = rate_limiter.get_url(rpc_dict.get('method'))
    if isinstance(url,dict):
        url=url.get('url')
    response = postRpcRequest(url=url,**rpc_dict)
    return response



from typing import List, Dict, Any, Optional, Union
from .pubkey_utils import try_pubkey
from .account_key_utils import get_all_account_keys

def search_for_index(data,index_number,key):
    for index_data in data:
        if str(index_data.get(key)) == str(index_number):
            return index_data
        
def search_for_account_index(data,index_number):
    return search_for_index(data,index_number,'accountIndex')

def find_log_entry(needle: str, log_entries: List[str]) -> Optional[str]:
    for log_entry in log_entries:
        if needle in log_entry:
            return log_entry
    return None

def extract_lp_initialization_log_entry_info_from_log_entry(lp_log_entry: str) -> Dict[str, Union[int, float]]:
    lp_initialization_log_entry_info_start = lp_log_entry.find('{')
    return json.loads(fix_relaxed_json_in_lp_log_entry(lp_log_entry[lp_initialization_log_entry_info_start:]))

def fix_relaxed_json_in_lp_log_entry(relaxed_json: str) -> str:
    return relaxed_json.replace(r'([{,])\s*([a-zA-Z_][a-zA-Z0-9_]*)\s*:', r'\1"\2":')

def get_associations(event, txnData):
    accountKeys = get_all_account_keys(txnData)
    key_properties = {'associatedAccounts': ['accountIndex', 'accounts'], 'programId': ['programIdIndex']}
    new_accounts = {k: None for k in key_properties}
    for key, values in key_properties.items():
        for value in values:
            subjects = event.get(value)
            if value in event and subjects is not None:
                subjects = event.get(value)
                if isinstance(subjects, list):
                    event[key] = [convert_subject(sub, accountKeys) for sub in subjects]
                else:
                    event[key] = convert_subject(subjects, accountKeys)
                    
    return event

def convert_subject(sub, accountKeys):
    return sub if try_pubkey(sub) else accountKeys[int(sub)]
